"""
Converts markdown to html
"""
import re
import os
from typing import Any, List, Tuple, Optional, Dict
import markdown2

from .log import logger

CF_URL = re.compile(r"(?P<host>https?://[^/]+)/.*/(?P<page_id>\d+)")
IMAGE_PATTERN = re.compile(r"\!\[(?P<alt>.*)\]\((?P<path>[^:)]+)\)")


def md_to_html(md_file: str,
               add_info_panel: bool) -> Tuple[Any, Optional[str], Optional[str], List]:
    """Converts given md_file to html"""

    logger.debug("Converting MD to HTML")
    md = __get_file_contents(md_file)
    images = __get_images_from_file(md)

    html = markdown2.markdown_path(
        path=md_file,
        extras=[
            "metadata",
            "strike",
            "tables",
            "wiki-tables",
            "code-friendly",
            "fenced-code-blocks",
            "footnotes",
        ],
    )
    page_id_from_meta, url = __parse_confluence_url(html.metadata)
    if add_info_panel:
        html = __get_info_panel(md_file) + html

    md_file_dir = os.path.dirname(md_file)
    html = __rewrite_images(html, md_file_dir, images)
    return html, page_id_from_meta, url, images

def __parse_confluence_url(meta: Dict[str, str]) -> Tuple[Optional[str], Optional[str]]:
    """Parses Confluence page URL and returns page_id and host"""
    if "confluence-url" not in meta:
        return (None, None)

    url = meta["confluence-url"]

    logger.debug("Looking for host and page_id in %s url", url)
    cf_url = CF_URL.search(url)
    if cf_url:
        page_id = cf_url.group("page_id")
        host = cf_url.group("host")
        logger.debug("  found page_id `%s` and host `%s`", page_id, host)
        return (page_id, host)
    logger.debug("  no valid Confluence url found")
    return (None, None)

def __get_info_panel(md_file: str) -> str:
    """
    Returns str with html info page to be placed on a Confluence page
    if --add_info is added
    """
    return f"""
        <p><strong>Automatic content</strong> This page was generated automatically from
        <code>{md_file}</code> file.Do not edit it on Confluence.</p><hr />
    """

def __get_images_from_file(md: str) -> List:
    logger.debug("Getting list of images from MD file")
    images = []
    for image in IMAGE_PATTERN.finditer(md):
        path = image.group("path")
        images.append((image.group("alt"), path))
        logger.debug("  - found image %s", path)
    return images

def __rewrite_images(html: str,
                     md_file_dir: str,
                     images: List[Tuple[str, str]]
                     ) -> str:
    """Replaces <img> html tags with Confluence specific <ac:image> and uploads
        images as attachements"""
    for (alt, path) in images:
        rel_path = os.path.join(md_file_dir, path)
        if not os.path.isfile(rel_path):
            assert os.path.isfile(path), f"File `{path}` does not exist"
            logger.debug("File `%s` does not exist, using file relative to "
                            "current dir `%s`", rel_path, path)
            rel_path = path

        old = f'<img src="{path}" alt="{alt}" />'
        new = f'<ac:image> <ri:attachment ri:filename="{os.path.basename(rel_path)}" />' \
                '</ac:image>'
        if html.find(old) != -1:
            logger.debug("replace image tag `%s` with `%s`", old, new)
            html = html.replace(old, new)
        else:
            logger.warning("image tag `%s` not found in html", old)
    return html

def __get_file_contents(file: str) -> str:
    """Return file contents"""
    with open(file, "r", encoding="utf-8") as stream:
        return stream.read()



#!/usr/bin/env python3
"""
confluence.md
Push markdown files straight to a Confluence page.
"""

import sys
import argparse
from argparse import RawTextHelpFormatter

from .utils.log import logger, init_logger, headline
from .utils.confluencemd import ConfluenceMD

ACTIONS = {}

def register_action(function):
    """Register command line action function decorator"""
    ACTIONS[function.__name__] = function.__doc__
    def wrapper(args):
        headline(function.__doc__)
        function(args)
        headline("End " + function.__name__)
    return wrapper

def init_confluence(args):
    """Inits connections to Confluence"""
    return ConfluenceMD(username=args.user,
                        token=args.token,
                        password=args.password,
                        md_file=args.file.name,
                        url=args.url,
                        verify_ssl=(not args.no_verify_ssl),
                        add_meta=args.add_meta,
                        add_info_panel=args.add_info,
                        add_label=args.add_label,
                        convert_jira=args.convert_jira)

@register_action
def update(args):
    """Updates page content based on given page_id or metadata in Markdown file"""
    confluence = init_confluence(args)
    confluence.update_existing(args.page_id)

@register_action
def create(args):
    """Creates new page under given parent_id"""
    assert args.url, ("No --url parameter is provided, gave up")

    confluence = init_confluence(args)
    confluence.create_new(args.parent_id, args.title, args.overwrite)

def main():
    """Markdown to Confluence

Example workflow:

1/ Create a new page under --parent_id:

  $ confluence.md --user user@name.net --token 9a8dsadsh --url https://your-domain.atlassian.net \\
        create --file README.md --parent_id 182371 --title "new title" --add_meta

2/ The page is created and the file is decorated with ### 2. The page is created and the file is
  decorated with metadata:

  $ head -n 3 markdown.md
  ---
  confluence-url: https://your-domain.atlassian.net/wiki/spaces/SP/pages/18237182/new+title
  ---

3/ Performing an update does not require providing --page_id and --url:

  $ confluence.md --user user@name.net --token 9a8dsadsh update --file README.md

  Doing an update with --page_id and --url is still possible.

  Consider adding useful --add_info option.

To create Atlassian API Token go to:
  https://id.atlassian.com/manage-profile/security/api-tokens

While using Atlassian on-premise instance, you need to provide --password instead of
--token parameter.

Actions:
"""

    actions = list(ACTIONS.keys())

    description = main.__doc__
    for action in actions:
        description += f"  {action:10}\t\t{ACTIONS[action]}\n"

    parser = argparse.ArgumentParser(
        add_help=True,
        prog="confluence.md",
        formatter_class=RawTextHelpFormatter,
        description=description)

    auth_args = parser.add_argument_group('required auth parameters')
    auth_args.add_argument("-u", "--user",
                           action="store",
                           required=True,
                           help="Atlassian username/email")
    auth_args.add_argument("-l", "--url",
                           action="store",
                           required=False,
                           help="Atlassian instance URL")
    auth_args.add_argument("-n", "--no_verify_ssl",
                           action="store_true",
                           required=False,
                           default=False,
                           help="Don't verify SSL cert in on-prem instances")

    secret_args = auth_args.add_mutually_exclusive_group(required=True)
    secret_args.add_argument("-t", "--token",
                             action="store",
                             required=False,
                             help="Atlassian API token (used in cloud instances)")
    secret_args.add_argument("-p", "--password",
                             action="store",
                             required=False,
                             help="Atlassian password (used in on-prem instances)")


    create_args = parser.add_argument_group('create page parameters')
    create_args.add_argument("--parent_id",
                             action="store",
                             help="define parent page id while creating a new page")
    create_args.add_argument("--title",
                             action="store",
                             help="define page title while creating a new page")
    create_args.add_argument("--overwrite",
                             action="store_true",
                             help="force overwrite if page with this title already exists")

    update_args = parser.add_argument_group('update page arguments')
    update_args.add_argument("--page_id",
                             action="store",
                             help="define (or override) page id while updating a page")

    parser.add_argument("--file",
                        action="store",
                        type=argparse.FileType('r'),
                        required=True,
                        help="input markdown file to process")

    parser.add_argument("--add_meta", action="store_true",
                        help="adds metadata to .md file for easy editing")
    parser.add_argument("--add_info", action="store_true",
                        help="adds info panel **automatic content** "
                            "do not edit on top of the page")
    parser.add_argument("--add_label", action="store",
                        help="adds label to page")
    parser.add_argument("--convert_jira",
                        action="store_true",
                        default=False,
                        help="convert all Jira links to issue snippets "
                            "(either short [KEY-ID] format or full URL)")

    parser.add_argument("-v", "--verbose",
                        action="store_true",
                        help="verbose mode")
    parser.add_argument("-q", "--quiet",
                        action="store_true",
                        help="quiet mode")
    parser.add_argument("action",
                        help="Action to run",
                        choices=ACTIONS)

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    init_logger(args)

    try:
        globals()[args.action](args)
    # pylint: disable=broad-exception-caught
    except (RuntimeError, AssertionError, Exception) as error:
        logger.error(error)
        if args.verbose:
            raise error
        quit(-1)


if __name__ == "__main__":
    main()



"""
Confluence to Markdown utility
"""
import os
import re
from  urllib import parse
from typing import List, Tuple
import requests

import atlassian

from .log import logger
from .md2html import md_to_html

ISSUE_PATTERN_KEY = re.compile(r"\[(?P<key>\w[\w\d]*-\d+)\]")
ISSUE_PATTERN_URL = re.compile(r"[^\"](?P<url>(?P<domain>https:\/\/\w+\.atlassian\.net)"
                               r"\/browse\/(?P<key>\w[\w\d]*-\d+))")

class ConfluenceMD(atlassian.Confluence):
    """Confluence to Markdown utility class"""
    jira_url:str = None
    conf_url:str = None

    def __init__(
        self,
        username: str,
        md_file: str,
        token: str = '',
        password: str = '',
        url: str = None,
        verify_ssl: bool = True,
        add_meta: bool = False,
        add_info_panel: bool = False,
        add_label: str = None,
        convert_jira: bool = True
    ) -> None:
        if url:
            self.jira_url = parse.urljoin(url, '/')
            self.conf_url = parse.urljoin(url, '/wiki/')

        super().__init__(
            url=self.conf_url or "",
            username=username,
            password=(password or token),
            verify_ssl=verify_ssl,
            cloud=bool(token),
            token=token,
        )

        self.license = self.__init_jira(
                url=self.jira_url,
                username=username,
                password=(password or token),
                verify_ssl=verify_ssl,
                token=token
            )
        self.md_file = md_file
        self.add_meta = add_meta
        self.add_info_panel = add_info_panel
        self.add_label = add_label
        self.md_file_dir = os.path.dirname(md_file)
        self.convert_jira = convert_jira

    def __init_jira(self,
                    url: str,
                    username: str,
                    password: str,
                    verify_ssl: bool,
                    token: str) -> bool:
        self.jira = atlassian.Jira(
                url=url,
                username=username,
                password=(password or token),
                verify_ssl=verify_ssl,
                cloud=bool(token),
                token=token
            )
        try:
            uri = parse.urljoin(url, 'wiki/rest/atlassian-connect/1/addons/secure-markdown-for-confluence')
            res = requests.get(uri, auth=(username, password or token), timeout=30)
            license_obj = res.json()
            if res.status_code != 200:
                return False
            return bool (license_obj["license"]["active"])
        # pylint: disable=broad-exception-caught
        except (RuntimeError, AssertionError, Exception):
            return False

    def update_existing(self, page_id: str = None) -> int:
        """Updates an existing page by given page_id"""
        logger.debug("Updating page `%s` based on `md_file` file", page_id)
        html, page_id_from_meta, url, images = md_to_html(self.md_file, self.add_info_panel)
        if self.conf_url is None:
            logger.debug("Using URL (%s) from `%s` file", url, self.md_file)
            assert url, (
                f"Can't update page without url given either by "
                f"`--url` parameter or via `confluence-url` tag in `{self.md_file}` file`"
            )
            self.jira_url = parse.urljoin(url, '/')
            self.conf_url = parse.urljoin(url, '/wiki/')
            self.url = self.conf_url # to satisfy parent class

        html = self.__rewrite_issues(html)
        if page_id is None:
            logger.debug("Using `page_id` from `%s` file", self.md_file)
            page_id = page_id_from_meta

        self.__attach_images(page_id, images)

        assert page_id, (
            f"Can't update page without page_id given either by "
            f"`--page_id` parameter or via `confluence-url` tag in `{self.md_file}` file"
        )


        title = self.__get_page_title_by_id(page_id)

        logger.debug("Updating page_id `%s` titled `%s`", page_id, title)
        response = self.update_page(
            page_id,
            title,
            html,
            parent_id=None,
            type="page",
            representation="storage",
            minor_edit=True,
        )

        if self.add_meta:
            confluence_url = ConfluenceMD.__get_link_from_response(response)
            self.__add_meta_to_file(confluence_url)

        if self.add_label:
            self.__add_label_to_page(page_id)

        return page_id

    def create_new(self, parent_id: str, title: str, overwrite: bool) -> int:
        """Creates a new page under give parent_id"""
        assert title, "Provide a title for a newly created page"
        assert parent_id, "Provide parent_id for a newly created page"
        space = self.get_page_space(parent_id)

        page_id = None
        if self.page_exists(space, title):
            page_id = self.get_page_id(space, title)
            assert overwrite, (
                f"Page titled `{title}` already exists in "
                f"the `{space}` space. Use `--overwrite` to force it."
            )

        html, page_id_from_meta, _url, images = md_to_html(self.md_file, self.add_info_panel)
        html = self.__rewrite_issues(html)
        assert not page_id_from_meta or overwrite, (
            f"Metadata pointing to an existing page "
            f"id `{page_id_from_meta}` present in the given markdown file. "
            f"Use `--overwrite` to force it."
        )

        overwrite_id = page_id if page_id else page_id_from_meta

        if overwrite_id:
            logger.debug(
                "Overwriting existing page `%s` based on `%s` file", title, self.md_file
            )
            response = self.update_page(
                overwrite_id,
                title,
                html,
                parent_id=None,
                type="page",
                representation="storage",
                minor_edit=True,
            )
        else:
            logger.debug("Creating new page `%s` based on `%s` file", title, self.md_file)
            response = atlassian.Confluence.create_page(
                self,
                space,
                title,
                body=html,
                parent_id=parent_id,
                type="page",
                representation="storage",
                editor="v2",
            )
            if images:
                logger.debug("Uploading images to newly created page")
                page_id = ConfluenceMD.__get_page_id_from_response(response)
                self.__attach_images(page_id, images)

        confluence_url = ConfluenceMD.__get_link_from_response(response)
        logger.debug(
            "%s %s", 'Page overwritten' if overwrite_id else 'New page created', confluence_url
        )

        self.__add_meta_to_file(confluence_url)

        page_id = ConfluenceMD.__get_page_id_from_response(response)
        self.__add_label_to_page(page_id)
        return page_id

    def __rewrite_issues(self, html):
        if self.convert_jira:
            logger.debug("Replacing [ISSUE-KEY] with html links")

        issues = []
        for issue in ISSUE_PATTERN_KEY.finditer(html):
            issues.append((issue.group(), issue.group("key")))
        for issue in ISSUE_PATTERN_URL.finditer(html):
            if self.jira_url.startswith(issue.group("domain")):
                issues.append((issue.group("url"), issue.group("key")))
            else:
                if self.convert_jira:
                    logger.info("Ignoring %s - domain mismatch (%s != %s)",
                                issue.group(),
                                issue.group("domain"),
                                self.jira_url)

        if issues and not self.convert_jira:
            (replace, key) = issues[0]
            logger.info("Use `--convert_jira` to replace %i Jira link(s) (such as %s) "
                        "with issue snippets - KEY: summary [status]",
                        len(issues), key)
            return html

        if not self.license and self.convert_jira:
            logger.warning("To expand Jira issues, you must have „Secure Markdown for Confluence” installed (free for small business)")
            logger.warning("Cloud: https://marketplace.atlassian.com/plugins/secure-markdown-for-confluence")
            logger.warning("Server/data-center: https://dirtyagile.atlassian.net/wiki/x/AQACR")
            return html

        for (replace, key) in issues:
            logger.debug("  - [%s] with html link", key)
            (summary, status, issuetypeurl) = self.__get_jira_issue(key)
            if summary:
                re_replace = re.compile(r"([^\"])" + re.escape(replace))
                html = re.sub(re_replace,
                    f"\\1<a href=\"{self.jira_url}browse/{key}\"><ac:image>"
                    f"<ri:url ri:value=\"{issuetypeurl}\" />"
                    f"</ac:image> {key}: {summary} [{status}]</a>",
                    html)
        return html

    def __get_jira_issue(self, key: str) -> tuple:
        try:
            issue = self.jira.issue(key)
            summary = issue['fields']['summary']
            status = issue['fields']['status']['name']
            issuetypeurl = issue['fields']['issuetype']['iconUrl']
            return (summary, status, issuetypeurl)
        # pylint: disable=broad-exception-caught
        except (RuntimeError, AssertionError, Exception) as error:
            logger.info("Unable to convert %s to Jira link: %s", key, error)
            return (None, None, None)

    def __attach_images(
            self, page_id: str, images: List[Tuple[str, str]]
    ) -> None:
        """Replaces <img> html tags with Confluence specific <ac:image> and uploads
           images as attachements"""
        logger.info("__attach_images %s", page_id)
        for (_alt, path) in images:
            rel_path = os.path.join(self.md_file_dir, path)
            if not os.path.isfile(rel_path):
                assert os.path.isfile(path), f"File `{path}` does not exist"
                logger.warning("File `%s` does not exist, using file relative "
                               "to current dir `%s`", rel_path, path)
                rel_path = path

            logger.debug("register image file `%s`", rel_path)
            self.attach_file(filename=rel_path, page_id=page_id)

    @staticmethod
    def __get_link_from_response(response) -> str:
        """Returns URL to page from Confluence API response"""
        return response["_links"]["base"] + response["_links"]["webui"]

    @staticmethod
    def __get_page_id_from_response(response) -> str:
        """Returns page_id from Confluence API response"""
        return response["id"]

    def __add_meta_to_file(self, confluence_url: str) -> None:
        """Decorates markdown file with metadata in comments"""
        if not self.add_meta:
            return

        markdown = ConfluenceMD.__get_file_contents(self.md_file)
        markdown = ("---\n" f"confluence-url: {confluence_url}\n" "---\n") + markdown

        with open(self.md_file, "w", encoding="utf-8") as stream:
            stream.write(markdown)

    def __get_page_title_by_id(self, page_id: str) -> str:
        """Returns page title by given page_id"""
        logger.debug("Getting page title from page id `%s`", page_id)
        page = self.get_page_by_id(page_id)
        assert "title" in page, f"Expected page-object while getting page by id, got {page}"
        return page["title"]

    @staticmethod
    def __get_file_contents(file: str) -> str:
        """Return file contents"""
        with open(file, "r", encoding="utf-8") as stream:
            return stream.read()


    def __add_label_to_page(self, page_id: str) -> None:
        """Selfdescriptive"""
        if not self.add_label:
            return
        self.set_page_label(page_id, self.add_label)
